print("import: 'behave'")
import behave

print("import: 'behave.compat'")
import behave.compat

print("import: 'behave.formatter'")
import behave.formatter

print("import: 'behave.reporter'")
import behave.reporter

